package com.falcon.core;

import com.falcon.core.binding.BindingConfig;
import com.falcon.core.binding.beans.ArgType;
import com.falcon.core.binding.beans.Binding.DataTransfer;
import com.falcon.core.binding.beans.Method;
import com.falcon.core.binding.beans.Method.Argument;
import com.falcon.core.exception.FalconCoreException;
import com.falcon.core.exception.FalconRunnerException;
import com.falcon.core.exception.NoDataTransferException;
import com.falcon.core.exception.NoOutObjectEception;
import com.falcon.core.util.ClassInstanceWrapper;
import com.falcon.core.util.ObjectInstanceStrategy;
import com.falcon.core.util.OutObjectHelper;
import org.apache.log4j.Logger;

import java.lang.reflect.Array;
import java.lang.reflect.InvocationTargetException;
import java.util.*;

/**
 * This class implements interface TwineRunner and has method to retrive object
 * with depenancy injected. <br>
 *
 * @author Charlie Date: 02-Apr-2006 Time: 18:33:37
 */
public class FalconAssembler implements FalconRunner {

    private static FalconAssembler _impIns = new FalconAssembler();
    private Logger log = Logger.getLogger(this.getClass().getName());
    private Map<String, Object> singletonMap = new HashMap<String, Object>();
    private static BindingConfig currentConfig;

    private FalconAssembler() {
    }

    public static FalconAssembler getInstance(BindingConfig conf) {
        currentConfig = conf;
        return _impIns;
    }

    /**
     * This method implements the interface. It returns the object that has
     * beening requested <br>
     *
     * @throws FalconRunnerException
     * @throws NoDataTransferException
     */
    public final Object getObject(String dtName) throws FalconCoreException {
        log.info("~~~~~~~~~~~~~~~~~~~~~~~GET OBJECT " + dtName + "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
        return bindDependancyManager(dtName, currentConfig, null, false);
    }

    /**
     * This methodimplements the interface method and allows getting of a object
     * where its dependancies are inject externally created object
     */
    public Object getObject(String dtName, Set<? extends Object> startObject)
            throws FalconCoreException {
        log.info("~~~~~~~~~~~~~~~~~~~~~~~GET OBJECT with StartObject " + dtName + "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");

        return bindDependancyManager(dtName, currentConfig, startObject, false);
    }

    /**
     * THis method implments the Runner interface and allows getting of object
     * as a 'By-product' manner.
     */
    // TODO: need ot enhance this method so it can handle out object method to
    // take in arguments.
    /*
      * Is the way that the out object being reference a little redendant since
      * no binding is nessary? so does it require a data transfer tag or just a
      * id is fine? However using dtRef could be a little more consistant?
      */
    public Object getObjectOut(String objectId) throws FalconCoreException {

        OutObjectHelper out = currentConfig.getOutPutDependantDt(objectId);

        if (out == null) {
            throw new NoOutObjectEception(objectId + " do not exist as a out object");
        } else {
            try {
                Object outObj = bindDependancyManager(out.getFromDt().getName(), currentConfig, null, true);
                Class outputClass = outObj.getClass();
                java.lang.reflect.Method methToCall = outputClass.getMethod(out.getFromMethod(), new Class[0]);
                return methToCall.invoke(outObj, new Object[0]);

            } catch (SecurityException e) {
                e.printStackTrace();
            } catch (NoSuchMethodException e) {
                e.printStackTrace();
            } catch (IllegalArgumentException e) {
                e.printStackTrace();
            } catch (IllegalAccessException e) {
                e.printStackTrace();
            } catch (InvocationTargetException e) {
                e.printStackTrace();
            }
        }

        return null;
    }

    public BindingConfig getCurrentBindingConfig() {

        return currentConfig;
    }

    /**
     * Overiding the current associated binding config. WARNING: This is NOT thread safe.
     *
     * @param conf
     */
    public void overrideCurrentBindingConfig(BindingConfig conf) {
        currentConfig = conf;
    }

    /**
     * This method acts as a manager of getting all the dependancy processed and
     * finally process the requested startObjects. All processed dependant startObjects are
     * put into a HashMap <br>
     *
     * @param dtName
     * @param conf
     * @return
     * @throws NoDataTransferException
     */
    protected final Object bindDependancyManager(String dtName, BindingConfig conf, Set<? extends Object> startObjects,
                                                 boolean out) throws NoDataTransferException, FalconRunnerException {

        List<DataTransfer> depDt = createOrderedDependantList(dtName, conf);
        Map<String, Object> depObjMap = new HashMap<String, Object>();
        ClassInstanceWrapper wrapper;

        if (startObjects != null) {
            log.debug("start objects are not null adding to map");
            // load start startObjects into map by <key=class name, value=startObjects in
            // set>
            for (Object singleObj : startObjects) {
                log.debug(singleObj.getClass().getName() + " added to map");
                depObjMap.put(singleObj.getClass().getName(), singleObj);
            }
        }

        for (DataTransfer dt : depDt) {
            if (!dt.isSingleton()) {
                wrapper = ObjectInstanceStrategy.getObjectInstance(dt, null, depObjMap);
            } else {
                if (singletonMap.containsKey(dt.getName())) {
                    wrapper = new ClassInstanceWrapper();
                    wrapper.setInstanceClass(singletonMap.get(dt.getName()).getClass());
                    wrapper.setInstanceObj(singletonMap.get(dt.getName()));
                } else {
                    wrapper = ObjectInstanceStrategy.getObjectInstance(dt, null, depObjMap);
                    singletonMap.put(dt.getName(), wrapper.getInstanceObj());
                }
            }

            depObjMap.put(dt.getName(), bindDependantObject(wrapper, dt, depObjMap));
        }

        if (out) {
            return getRequestedObject(conf.getDtByName(dtName), depObjMap);

        } else {
            return getRequestedObject(conf.getDtByName(dtName), depObjMap);
        }
    }

    /**
     * x Method that get the requested object and add all dependent object from
     * mao builed
     *
     * @param reqDt
     * @param depeantObjs
     * @return
     * @throws FalconRunnerException
     */
    protected final Object getRequestedObject(DataTransfer reqDt, Map<String, Object> depeantObjs)
            throws FalconRunnerException {
        log.debug("===================getting requested Object " + reqDt.getName()
                + "======================================");
        ClassInstanceWrapper reqObjectWrapper = ObjectInstanceStrategy.getObjectInstance(reqDt, null, depeantObjs);

        return bindDependantObject(reqObjectWrapper, reqDt, depeantObjs);
    }

    protected final Object bindDependantObject(ClassInstanceWrapper rntInsWrap, DataTransfer dt,
                                               Map<String, Object> depObjMap) throws FalconRunnerException {

        log.debug("-----------------------processing dt " + dt.getName() + "-----------------------------------------");
        Class insClass = rntInsWrap.getInstanceClass();
        List<Argument> argsList = new ArrayList<Argument>();
        List<Class> argClassList = new ArrayList<Class>();
        List<Object> argObjList = new ArrayList<Object>();
        Object instanceObj = rntInsWrap.getInstanceObj();

        for (Method meth : dt.getMethod()) {
            argsList.addAll(meth.getArgument());

            for (Argument arg : argsList) {
                if (arg.getType() != ArgType.OUTPUT) {
                    argClassList.add(getArgClass(depObjMap, meth, arg));
                    argObjList.add(getArgObject(depObjMap, arg));
                }
            }


            invoke(meth.getName(), insClass, argClassList.toArray(new Class[argClassList.size()]), instanceObj, argObjList.toArray(), false);
            argClassList.clear();
            argObjList.clear();
            argsList.clear();
        }
        log.debug("-----------------------processing dt " + dt.getName()
                + " FINISHED-----------------------------------------");
        return instanceObj;
    }

    private Object invoke(String methName, Class insClass, Class[] argTypes, Object instanceObj, Object[] argObjects,
                          boolean out) throws FalconRunnerException {
        try {
            log.debug("calling method " + methName + " number of argments " + argTypes.length + " with arg data "
                    + argObjects);
            java.lang.reflect.Method invokMeth = insClass.getDeclaredMethod(methName, argTypes);
            if (out) {
                return invokMeth.invoke(instanceObj, argObjects);
            } else {
                invokMeth.invoke(instanceObj, argObjects);
            }

        } catch (SecurityException e) {

            log.error("There is a security exp calling method " + methName + " ", e.fillInStackTrace());
            throw new FalconRunnerException("There is a security exp calling method " + methName, e.fillInStackTrace());
        } catch (NoSuchMethodException e) {

            log.error("There is no such method for meth Definition " + methName + " ", e.fillInStackTrace());
            throw new FalconRunnerException("There is no such method for meth Definition " + methName, e
                    .fillInStackTrace());
        } catch (IllegalArgumentException e) {

            log.error("The arguments supplied threw a exception for method " + methName + " ", e.fillInStackTrace());
            throw new FalconRunnerException("The arguments supplied threw a exception for method " + methName, e
                    .fillInStackTrace());
        } catch (IllegalAccessException e) {

            log.error("There is a IllegalAccessException ", e.fillInStackTrace());
            throw new FalconRunnerException("There is a IllegalAccessException ", e.fillInStackTrace());
        } catch (InvocationTargetException e) {

            log.warn("Calling method " + methName + " threw a exception ");
            throw new FalconRunnerException("Calling method " + methName + " threw a exception ", e.getCause());
        }

        return null;
    }

    /**
     * private method that get the argument data from for that argument. Also
     * deal with array situations if the <argument array='true'.... <br>
     *
     * @param depObjMap
     * @param arg
     * @return
     * @throws FalconRunnerException
     */
    private Object getArgObject(Map<String, Object> depObjMap, Argument arg) throws FalconRunnerException {
        Object rntObject = null;

        if (!arg.isArray()) {
            if (arg.getStatic().size() > 0) {
                rntObject = DataTypeConvertor.convertDataTypeData(arg.getStatic().get(0), arg.getDataType());
            } else if (arg.getDtRef() != null) {

                if (depObjMap.get(arg.getDtRef()) != null)
                    rntObject = depObjMap.get(arg.getDtRef());
            } else if (arg.getRefClass() != null) {

                if (depObjMap.get(arg.getRefClass()) != null)
                    rntObject = depObjMap.get(arg.getRefClass());
            }
        } else {
            log.debug("arg is an array");
            List<Object> resultList = new ArrayList<Object>();
            if (arg.getStatic().size() > 0) {
                log.debug("gettting static data");
                for (String stat : arg.getStatic()) {
                    log.debug("static data added to list " + stat);
                    resultList.add(DataTypeConvertor.convertDataTypeData(stat, arg.getDataType()));
                }
                Object[] restAr = (Object[]) Array.newInstance(DataTypeConvertor.getDataTypeClass(arg.getDataType(),
                        false), resultList.size());
                log.debug("New ArrayClass len " + restAr.length + " of type " + restAr.getClass().getName());
                resultList.toArray(restAr);
                rntObject = restAr;
            } else if (arg.getDtRef() != null) {
                throw new FalconRunnerException("dtRef cannot be array");
            } else if (arg.getRefClass() != null) {
                rntObject = depObjMap.get(arg.getDtRef());
            }
        }

        return rntObject;
    }

    /**
     * private method that gets the class type of the that is defined in
     * <argument datatype=..... Also if dtref att is set it will get the class
     * of the depenant object.
     *
     * @param depObjMap
     * @param meth
     * @param arg
     * @return
     * @throws FalconRunnerException
     */
    private Class getArgClass(Map<String, Object> depObjMap, Method meth, Argument arg) throws FalconRunnerException {
        log.debug("getArgClass called");
        Class argClass = null;

        if (arg.getDataType() != null) {
            argClass = DataTypeConvertor.getDataTypeClass(arg.getDataType(), arg.isArray());

            if (argClass == null) {
                log.error("class not found in datatype set in argument tag");
                throw new FalconRunnerException("class not found in datatype set in argument :" + arg.getDataType());
            }
        } else if (depObjMap.get(arg.getDtRef()) != null) {

            argClass = getRefClass(depObjMap, arg.getDtRef(), arg.isArray());

        } else if (depObjMap.get(arg.getRefClass()) != null) {
            log.debug("ref class found " + arg.getRefClass());
            argClass = getRefClass(depObjMap, arg.getRefClass(), arg.isArray());

        } else {
            log.error("No dataType found and no dt ref Found");
            throw new FalconRunnerException("no data type found and no dtref found for meth " + meth.getName());
        }

        return argClass;
    }

    private Class getRefClass(Map<String, Object> depObjMap, String mapKey, boolean array) {
        if (array) {
            return DataTypeConvertor.getArrayClass(depObjMap.get(mapKey).getClass());
        } else {
            return depObjMap.get(mapKey).getClass();
        }

    }

    /**
     * gets the list of dataTransfer objects that are depeneant by the original
     * one. the list is then reviersed where the last one to be processed first.
     * <br>
     *
     * @param objectId
     * @param conf
     * @return
     * @throws NoDataTransferException
     */
    protected final List<DataTransfer> createOrderedDependantList(String objectId, BindingConfig conf)
            throws NoDataTransferException {
        log.debug("createOrderedDependantList called");
        List<DataTransfer> orderProcessList = new ArrayList<DataTransfer>();
        List<DataTransfer> depDts = conf.getDependentDataTransfer(objectId);
        List<DataTransfer> tempDt = new ArrayList<DataTransfer>();

        if (!depDts.isEmpty())
            orderProcessList.addAll(depDts);

        for (DataTransfer dt : depDts) {
            tempDt.addAll(conf.getDependentDataTransfer(dt.getName()));
            if (tempDt.size() > 0) {
                orderProcessList.addAll(tempDt);
            }
            tempDt.clear();
        }

        Collections.reverse(orderProcessList);

        return orderProcessList;
    }
}
