package com.falcon.core.util;

/**
 * Util class that parses a XML Document
 * <br>
 * Date: 16-Dec-2004
 * Within Package: com.falcon.core.util
 * @author Charles Lee
 */

import com.falcon.core.binding.beans.Falcon;
import com.falcon.core.exception.FalconCoreException;
import com.falcon.core.exception.FalconFileException;
import com.falcon.core.exception.FalconValidationException;
import org.apache.log4j.Logger;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.Vector;

public final class BeansMapper {
    private static Logger log = Logger.getLogger("com.falcon.core.util.BeansMapper");

    private static final String _jaxbContext = "com.falcon.core.binding.beans";

    private BeansMapper() {
    }

    /**
     * returns a list of fal Files <br>
     *
     * @param dir
     * @return {@link java.io.File} array
     * @throws FileNotFoundException
     */
    public static File[] getFiles(String dir) throws FileNotFoundException {
        Vector<File> v = new Vector<File>();
        File fileList = new File(dir);

        if (fileList.exists()) {
            File[] fList = fileList.listFiles();
            for (File aFList : fList) {
                log.debug("trying to match file name: " + aFList.getName());
                if (aFList.getName().matches(".*.fal") & aFList.isFile()) {
                    log.debug("this is a falcon file: " + aFList.getName());
                    v.add(aFList);
                    log.debug(aFList.getPath() + " added to list");
                } else {
                    log.warn(aFList.getName() + " is not a falcon file");
                }
            }
        } else {
            throw new FileNotFoundException("The dir name " + dir + " cannot be located");
        }

        return v.toArray(new File[v.size()]);
    }

    /**
     * public method to return a falcon object using JAXB binding function <br>
     *
     * @param falFile
     * @return {@link com.falcon.core.binding.beans.Falcon}
     */
    // TODO:change to jibx when it is ready
    public static Falcon getFalconObject(File falFile) throws FalconCoreException {
        Falcon unMarshObj = null;
        try {
            JAXBContext context = JAXBContext.newInstance(_jaxbContext);
            Unmarshaller unMarsh = context.createUnmarshaller();
            unMarshObj = (Falcon) unMarsh.unmarshal(falFile);
        } catch (JAXBException e) {
            throw new FalconValidationException("A Jaxb exception is thrown", e.getCause());
        }
        return unMarshObj;
    }

    public static File getFile(String falFile) throws FalconCoreException {
        File file = new File(falFile);
        if (file.getName().matches(".*.fal") & file.isFile()) {
            return file;
        } else {
            throw new FalconFileException("This is not a falcon file");
        }
    }
}
