package com.falcon.core.util;

import com.falcon.core.binding.beans.ObjectStory;
import com.falcon.core.exception.FalconJpaException;
import org.apache.log4j.Logger;

import javax.persistence.EntityManager;
import javax.persistence.Persistence;
import javax.persistence.Query;
import java.lang.reflect.InvocationTargetException;
import java.util.*;

public class JpaStrategyImp {
    private static Logger log = Logger
            .getLogger("com.twine.core.util.ObjectInstanceStrategy");

    private JpaStrategyImp() {
    }

    public static ClassInstanceWrapper getJpa(String persistenceUnit,
                                              String query, boolean isList, Map<String, Object> depObjMap,
                                              List<ObjectStory.JPA.Parameters> parameters) throws FalconJpaException {
        log.debug("classInstanceWwrapper called: persistenceUnit: " + persistenceUnit + " , query: "
                + query + " , isList: " + isList);

        ClassInstanceWrapper wrap = new ClassInstanceWrapper();
        EntityManager em = Persistence.createEntityManagerFactory(
                persistenceUnit).createEntityManager();
        log.debug("created EntityManager for unit name " + persistenceUnit);
        Query queryObj = em.createQuery(query);

        log.debug("no of parameters " + parameters.size());
        if (parameters.size() > 0) {
            log.debug("setting paramters");
            Map<String, Object> paramMap = getParamMap(getParamNames(query), parameters, depObjMap);
            log.debug("amount in paramMap: " + paramMap.size());
            for (String key : paramMap.keySet()) {
                queryObj.setParameter(key, paramMap.get(key));
            }
        }

        log.debug("created query " + query);
        if (isList) {
            log.debug("This query returns list");
            List queryResult = queryObj.getResultList();
            log.debug("got query Result " + queryResult.size());
            wrap.setInstanceClass(queryResult.getClass());
            wrap.setInstanceObj(queryResult);
        } else {
            log.debug("This query returns singleObject");
            Object queryResult = queryObj.getSingleResult();
            wrap.setInstanceClass(queryResult.getClass());
            wrap.setInstanceObj(queryResult);
        }
        return wrap;
    }

    private static Map<String, Object> getParamMap(List<String> paramNames,
                                                   List<ObjectStory.JPA.Parameters> parameters,
                                                   Map<String, Object> depObjMap) throws FalconJpaException {
        log.debug("paramNames: " + paramNames);
        Map<String, Object> paramMap = new HashMap<String, Object>(paramNames.size());
        List paramValues = getParamValues(parameters, depObjMap);

        for (int i = 0; i < paramNames.size(); i++) {
            if (paramValues.get(i) != null) {
                paramMap.put(paramNames.get(i), paramValues.get(i));
            }
        }

        return paramMap;
    }

    private static List getParamValues(List<ObjectStory.JPA.Parameters> parameters, Map<String, Object> depObjMap) throws FalconJpaException {
        List rntValues = new ArrayList();

        for (ObjectStory.JPA.Parameters params : parameters) {
            if (params.getClassName() == null & params.getDtRef() == null) {
                log.debug("getting param values NOT from dep objs");
                for (String paramValue : params.getParameter()) {
                    rntValues.add(paramValue);
                }
            } else if (params.getDtRef() != null) {
                log.debug("dtRef is not null: " + params.getDtRef());
                if (depObjMap.get(params.getDtRef()) != null) {
                    log.debug("found dtRef class: " + depObjMap.get(params.getDtRef()).getClass().getName());
                    Class refClass = depObjMap.get(params.getDtRef()).getClass();
                    for (String paramValue : params.getParameter()) {
                        try {
                            rntValues.add(refClass.getMethod(getBeanMethod(paramValue)).
                                    invoke(depObjMap.get(params.getDtRef())));
                        } catch (IllegalAccessException e) {
                            log.error("IllegalAccessException", e);
                            throw new FalconJpaException("IllegalAccessException", e);
                        } catch (InvocationTargetException e) {
                            log.error("InvocationTargetException", e);
                            throw new FalconJpaException("InvocationTargetException", e);
                        } catch (NoSuchMethodException e) {
                            log.error("NoSuchMethodException", e);
                            throw new FalconJpaException("NoSuchMethodException", e);
                        }
                    }
                } else {
                    throw new FalconJpaException("Cannot find dtref '" + params.getDtRef() + "' from depObj");
                }
            }
        }

        return rntValues;
    }

    private static List<String> getParamNames(String query) {
        List<String> names = new ArrayList<String>();
        StringTokenizer queryTok = new StringTokenizer(query, " ");
        String tempTok = "";
        while (queryTok.hasMoreTokens()) {
            tempTok = queryTok.nextToken();
            if (tempTok.startsWith(":")) {
                names.add(tempTok.substring(1));
            }
        }

        return names;
    }

    /**
     * private method that splits the file name of the first letter. COnverting that to a upper case
     * then getting the rest of the field name and concat them together and add
     * 'get' to the begining.
     *
     * @param fieldName
     * @return
     */
    private static String getBeanMethod(String fieldName) {
        log.debug("converting field: " + fieldName);
        return "get" + fieldName.substring(0, 1).toUpperCase().concat(fieldName.substring(1));
    }


}
