package com.falcon.core.util;

import com.falcon.core.binding.beans.Binding.DataTransfer;
import com.falcon.core.binding.beans.DataTransferType;
import com.falcon.core.binding.beans.ObjectStory;
import com.falcon.core.exception.FalconJpaException;
import org.apache.log4j.Logger;

import javax.naming.OperationNotSupportedException;
import java.util.List;
import java.util.Map;
import java.util.Properties;

public final class ObjectInstanceStrategy {

    private ObjectInstanceStrategy() {
    }

    private static Logger log = Logger
            .getLogger("com.twine.core.util.ObjectInstanceStrategy");

    public static ClassInstanceWrapper getObjectInstance(DataTransfer dt,
                                                         Properties remoteProperties, Map<String, Object> depObjMap) {
        log.info("getting object instance for " + dt.getClassObject()
                + " type " + dt.getType());

        ClassInstanceWrapper rntObj = null;
        if (dt.getType() == DataTransferType.POJO) {
            rntObj = getPojo(dt.getClassObject());
        } else if (dt.getType() == DataTransferType.JPA) {
            ObjectStory.JPA jpa = dt.getObjectStory().getJPA();
            rntObj = getJpaPojo(jpa.getPersistenceUnitName(), jpa.getQuery(), jpa.isReturnAsList(), depObjMap,
                    dt.getObjectStory().getJPA().getParameters());
        }

        return rntObj;
    }

    /**
     * Method to get POJO class information.
     *
     * @param className
     * @return
     */
    protected static ClassInstanceWrapper getPojo(String className) {
        ClassInstanceWrapper rntObj = new ClassInstanceWrapper();
        try {
            Class cl = Class.forName(className);
            rntObj.setInstanceClass(cl);
            rntObj.setInstanceObj(cl.newInstance());
        } catch (ClassNotFoundException e) {
            log.error("Problem createing class for class " + className, e);
        } catch (InstantiationException e) {
            log.error("problem create class instance", e);
        } catch (IllegalAccessException e) {
            log.error("problem accessing default constructor", e);
        }
        return rntObj;
    }

    /**
     * Method to get EJB Object TODO: not yet implmented
     *
     * @param className
     * @param remoteProps
     * @return
     */
    protected static ClassInstanceWrapper getEjb(String className,
                                                 Properties remoteProps) {
        try {
            throw new OperationNotSupportedException(
                    "Function not yet implmeneted");
        } catch (OperationNotSupportedException e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * Method to get JPA based Oject. At the moment this method only
     * allow retival of Objects it doesn't set any to DB
     *
     * @param persistenceUnit
     * @param query
     * @param isList
     * @param depObjMap
     * @param parameters
     * @return
     */
    protected static ClassInstanceWrapper getJpaPojo(
            String persistenceUnit, String query, boolean isList, Map<String, Object> depObjMap,
            List<ObjectStory.JPA.Parameters> parameters) {
        try {
            return JpaStrategyImp.getJpa(persistenceUnit, query, isList, depObjMap, parameters);
        } catch (FalconJpaException e) {
            log.error("problem with JPA related operateion", e);
        }

        return null;
    }
}
